

/**
 * This is the Control Library for the Kater Project
 *
 * (c) 2011
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * @author		Christoph Steinlehner http://angstproject.com
 * @modified	01/23/2011
 * @version		0.1.1
 */

package katerlib;

import java.awt.event.KeyEvent;
import java.io.InputStream;
import java.io.PrintWriter;
import java.lang.reflect.Method;
import java.util.Properties;
import java.util.Vector;

import TUIO.TuioObject;
import TUIO.TuioProcessing;
import processing.core.*;

/**
 * 
 * @example KaterExample
 * 
 * (the tag @example followed by the name of an example included in folder 'examples' will
 * automatically include the example in the javadoc.)
 *
 */

public class KaterKontrol implements IKaterEventListener{
	
	
	/*
	 *  Config Parameter
	 */
	private int cursorsize = 15;
	private int objectsize = 90;
	private int sensorradius = 45;
	private int sensorspotsize = 70;
	private int angleblur = 20;
	private int distanceblur = 40;
	private int xoffset = 0;
	private int yoffset = 0;
	private int angleOffset = 90;
	private int innerRadius = 90;
	private Boolean debug = false;
	private int debugState=0;
	
	
	// Java-properties for configuration
	private Properties config = new Properties();
	
	// myParent is a reference to the parent sketch
	private PApplet myParent;
	
	// tuioClient is a reference to the tuioClient in the parent sketch
	private TuioProcessing tuio;
	
	// Array of Fiducial IDs
    private int[] katerIds;
    
    // Arrays for the Kater-Management
    private Vector<Kater> katerList = new Vector<Kater>();
    private Vector<Kater> activeKater = new Vector<Kater>();
    
    
    // Methods for EventSystem
    Method katerStarted;
    Method katerFinished;
    
    // Version of the Library
	public final static String VERSION = "0.1.1";

	/**
	 * The KaterKontrol Constructor which needs a Reference to PApplet, TUIO and an array with Fiducial IDs 
	 * 
	 * @example KaterExample
	 * @param theParent	Reference to the PApplet (usually this)
	 * @param tuioClient Reference to the TUIOProcessing Client
	 * @param katerTuioIds	An Array of ints with the Fiducial IDs
	 * 
	 */
	public KaterKontrol(PApplet theParent, TuioProcessing tuioClient, int[] katerTuioIds) {
		myParent = theParent;
		this.tuio = tuioClient;
		this.katerIds=katerTuioIds;
		myParent.registerDispose(this);
		myParent.registerDraw(this);
		myParent.registerPre(this);
		myParent.registerKeyEvent(this);
		checkImplementations();
		readConfig();
	
		createKater();
	}
	/**
	 * 
	 * Checks if the EventDispatch Methods are implemented
	 */
	private void checkImplementations(){		
			    try { 
			    	katerFinished = myParent.getClass().getMethod("katerFinished", new Class[] { Kater.class }); 
			    }catch (Exception e) {
			     		System.out.println("katerLib: katerFinished Method ('katerFinished(Kater k)') isn't properly implemented");
			     		katerFinished = null;
			     		}
			    try { 
				   	katerStarted = myParent.getClass().getMethod("katerStarted", new Class[] { Kater.class }); 
			    	}catch (Exception e) {
			    		System.out.println("katerLib: katerStarted Method ('katerStarted(Kater k)') isn't properly implemented");
			    		katerStarted = null;
		     			}
		
	}
	/**
	 * Reads the config.properties from the data folder or creates one if there isn't any
	 * 
	 */
	private void readConfig(){
		// load config into Input else create file
		InputStream configInput;
		if(myParent.sketchFile("data/config.properties").canRead()){
			configInput = myParent.createInput("config.properties");
		}
		else{
			  System.err.println("katerLib: Config File can't be found, so I create one.");
			  PrintWriter output = myParent.createWriter("data/config.properties");
			  output.println("cursorsize = "+cursorsize);
			  output.println("objectsize = "+objectsize);
			  output.println("sensorradius = "+sensorradius);
			  output.println("sensorspotsize = "+sensorspotsize);
			  output.println("angleblur = "+angleblur);
			  output.println("distanceblur = "+distanceblur);
			  output.println("xoffset = "+xoffset);
			  output.println("yoffset = "+yoffset);
			  output.println("angleOffset = "+angleOffset);
			  output.println("innerRadius = "+innerRadius);
			  output.flush();
			  output.close();
			  configInput = myParent.createInput("config.properties");
		}
		// load config into variables
		try {
            config.load(configInput);   
            cursorsize = Integer.parseInt(config.getProperty("cursorsize"));
            objectsize = Integer.parseInt(config.getProperty("objectsize"));
            sensorradius = Integer.parseInt(config.getProperty("sensorradius")); 
            sensorspotsize = Integer.parseInt(config.getProperty("sensorspotsize")); 
            angleblur = Integer.parseInt(config.getProperty("angleblur"));
            distanceblur = Integer.parseInt(config.getProperty("distanceblur"));
            xoffset = Integer.parseInt(config.getProperty("xoffset"));
            yoffset = Integer.parseInt(config.getProperty("yoffset"));
            angleOffset = Integer.parseInt(config.getProperty("angleOffset"));
            innerRadius = Integer.parseInt(config.getProperty("innerRadius"));
             }catch(Exception e)
             {
                System.err.println("katerLib: There is something wrong with your config file, please fix it!");
             }
	}
	/**
	 * create the initial Kater Objects
	 */
	private void createKater(){
		 for (int i=0; i<katerIds.length;i++){
			 Kater tKater = new Kater(myParent, tuio, katerIds[i]);
			 tKater.setValues(sensorradius, sensorspotsize, angleblur, distanceblur, xoffset, yoffset, angleOffset, innerRadius);
			 tKater.addActionListener(this);
			 katerList.add(tKater);
		    }
	}
	private void setKaterValues(){
		for (int i = 0; i < katerList.size(); i++) {
			katerList.get(i).setValues(sensorradius, sensorspotsize, angleblur, distanceblur, xoffset, yoffset, angleOffset, innerRadius);
		}
	}
	/**
	 * 
	 * Kater will be update in the Processing Pre Cycle
	 */
	public void pre(){
			 Vector tuioObjectList = tuio.getTuioObjects();
			    for (int i = 0; i < tuioObjectList.size(); i++) {
			        TuioObject tobj = (TuioObject) tuioObjectList.elementAt(i);
			        for (int j=0; j<katerIds.length;j++){
			            if (tobj.getSymbolID()==katerIds[j]){
			            	katerList.get(j).update(tobj.getScreenX(myParent.width),tobj.getScreenY(myParent.height), tobj.getAngleDegrees());
			            }
			        }
			    }
	}
	/**
	 * 
	 * Kater will be drawn in the Processing Draw Cycle
	 */
	public void draw(){
		for (int i = 0; i < katerList.size(); i++) {
			katerList.get(i).draw();
		}
		if(debug){
			drawDebug();
		}
	}
	
	private void drawDebug(){
		myParent.pushMatrix();
		myParent.noFill();
		myParent.stroke(255);
		myParent.strokeWeight(1);
		myParent.rect(20,20,200,24);
		myParent.text("Framerate: "+ myParent.frameRate, 24, 38);
        
        if (debugState==0) myParent.stroke(100,100,255);
        else myParent.stroke(70);
        myParent.rect(20,50,200,24);
        myParent.text("SensorRadius: "+ sensorradius, 24, 68);
        if (debugState==1) myParent.stroke(100,100,255);
        else myParent.stroke(70);
        myParent.rect(20,80,200,24);
        myParent.text("SensorSpotsize: "+ sensorspotsize, 24, 98);
        if (debugState==2) myParent.stroke(100,100,255);
        else myParent.stroke(70);
        myParent.rect(20,110,200,24);
        myParent.text("AngleBlur: "+ angleblur, 24, 128);
        if (debugState==3) myParent.stroke(100,100,255);
        else myParent.stroke(70);
        myParent.rect(20,140,200,24);
        myParent.text("DistanceBlur: "+ distanceblur, 24, 158);
        if (debugState==4) myParent.stroke(100,100,255);
        else myParent.stroke(70);
        myParent.rect(20,170,200,24);
        myParent.text("Xoffset: "+ xoffset, 24, 188);
        if (debugState==5) myParent.stroke(100,100,255);
        else myParent.stroke(70);
        myParent.rect(20,200,200,24);
        myParent.text("Yoffset: "+ yoffset, 24, 218);
        if (debugState==6) myParent.stroke(100,100,255);
        else myParent.stroke(70);
        myParent.rect(20,230,200,24);
        myParent.text("AngleOffset: "+ angleOffset, 24, 248);
        myParent.popMatrix();
	}
	/**
	 * send a Kater by its Object to a Coordinate
	 * 
	 * @param kater	Reference to the Kater Object
	 * @param theX	X Coordinate
	 * @param theY	Y Coordinate
	 */
	public void KaterGoTo(Kater kater, float theX, float theY){
	    kater.goTo(theX, theY);
	    activeKater.add(kater);
	}
	/**
	 * send a Kater by its Fiducial ID to a Coordinate
	 * 
	 * @param katerId	Fiducial ID of the Kater
	 * @param theX	X Coordinate
	 * @param theY	Y Coordinate
	 */
	public void KaterIdGoTo(int katerId, float theX, float theY){
		getKaterById(katerId).goTo(theX, theY);
	}
	/**
	 * 
	 * Returns a Kater Object by its Fiducial ID
	 * 
	 * @param katerId Fiducial ID of the Kater 
	 * @return	the Kater Object
	 */
	public Kater getKaterById(int katerId){
		for (int i = 0; i < katerList.size(); i++) {
			if(katerList.get(i).getID()==katerId){
				return katerList.get(i);
			}
		}
		System.err.println("No Kater with this ID found");
		return null;
	}
	

	public void onActionStateChange(IKaterEventDispatcher dispatcher, EKaterEventState theState) {
		switch (theState) {
		case finished:
			if (katerFinished != null) {
			try {
				katerFinished.invoke(myParent, new Object[] { dispatcher });
			    } catch (Exception e) {
			      System.err.println("Disabling katerFinished() because of an error.");
			      e.printStackTrace();
			      katerFinished = null;}
			 }
			break;
		case started:
			if (katerStarted != null) {
			try {
				katerStarted.invoke(myParent, new Object[] { dispatcher });
			    } catch (Exception e) {
			      System.err.println("Disabling katerStarted() because of an error.");
			      e.printStackTrace();
			      katerStarted = null;}
			}
			break;
		default:
			break;
		}
		
	}
	
	/**
	 * return the version of the library.
	 * 
	 * @return String Version of the Library
	 */
	public static String version() {
		return VERSION;
	}
	/**
	 * 
	 *  Should only called by Processing. Empties the Vectors and end connections to Processing and TUIO
	 */
	public void dispose(){
		katerList=null;
		activeKater=null;
		tuio=null;
		myParent=null;
	}
	/**
	 * 
	 * 
	 * @param e the KeyEvent from @myParent
	 */
	public void keyEvent(KeyEvent e){
		// navigation for debug menu
		
		
		int keyCode = e.getKeyCode();
		if(debug){
			System.out.println("KeyEvent: "+e);
			 if (keyCode == PConstants.UP) {
                 if (debugState>0) debugState--;
             }
			 if (keyCode == PConstants.DOWN) {
                 if (debugState<6) debugState++;
             }
			 if (keyCode == PConstants.LEFT) {
                 switch (debugState) {
                     case 0: sensorradius--;
                         break;
                     case 1: sensorspotsize--;
                         break;
                     case 2: angleblur--;
                         break;
                     case 3: distanceblur--;
                         break;
                     case 4: xoffset--;
                         break;
                     case 5: yoffset--;
                         break;
                     case 6: angleOffset--;
                         break;
                 }
                 
             }
             if (keyCode == PConstants.RIGHT) {
                 switch (debugState) {
                     case 0: sensorradius++;
                         break;
                     case 1: sensorspotsize++;
                         break;
                     case 2: angleblur++;
                         break;
                     case 3: distanceblur++;
                         break;
                     case 4: xoffset++;
                         break;
                     case 5: yoffset++;
                         break;
                     case 6: angleOffset++;
                         break;
                 }
             }
             setKaterValues();
		}
		if(e.getKeyChar()=='d'){
			toggleDebug();
			for (int i = 0; i < katerList.size(); i++) {
				katerList.get(i).toggleDebug();
			}
		}
	}
	public Boolean getDebug() {
		return debug;
	}
	public void setDebug(Boolean debug) {
		this.debug = debug;
	}
	public void toggleDebug(){
		debug=!debug;
	}
}
